package eu.svjatoslav.inspector.java.commandline;

import eu.svjatoslav.inspector.java.structure.ClassGraph;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.ArrayList;
import java.util.List;
import java.util.jar.JarEntry;
import java.util.jar.JarInputStream;

/**
 * This class acts as a commandline interface for JavaInspect.
 */
public class Main {
    public static void main(String[] args) throws IOException, ClassNotFoundException {
        CommandlineConfiguration configuration = new CommandlineConfiguration(args);
        if (!configuration.configurationOk) {
            configuration.parser.showHelp();
            System.exit(1);
        }

        if (configuration.showHelp.isSpecified()) configuration.parser.showHelp();

        getClassGraph(configuration).generateGraph(configuration.graphName.getValue());

        if (configuration.isDebug())
            System.out.println("Graph ready.");
    }

    private static ClassGraph getClassGraph(CommandlineConfiguration configuration) throws IOException, ClassNotFoundException {
        List<File> jarFiles = configuration.jarFiles.getValue();

        URLClassLoader classLoader = new URLClassLoader(
                getClasspathUrls(jarFiles, configuration.classPaths.getValue()),
                configuration.getClass().getClassLoader());

        ClassGraph classGraph = new ClassGraph();

        if (configuration.targetDirectory.isSpecified())
            classGraph.setTargetDirectory(configuration.targetDirectory.getValue());

        if (configuration.targetImageType.isSpecified())
            classGraph.setTargetImageType(configuration.targetImageType.getValue());

        classGraph.setKeepDotFile(configuration.keepDotFile.getValue());

        if (configuration.rootClasses.isSpecified()) {
            // add only selected root classes
            for (String rootClass : configuration.rootClasses.getValue())
                attemptClassAdditionByName(classLoader, classGraph, configuration, rootClass);

        } else {
            // add all classes in the jar files to graph
            for (File jarFile : jarFiles)
                addJarToGraph(jarFile, classLoader, classGraph, configuration);
        }
        configuration.blacklistGlob.getValue().forEach(classGraph::blacklistClassGlob);
        configuration.whitelistGlob.getValue().forEach(classGraph::whitelistClassGlob);

        if (configuration.hideOrphanedClasses.getValue())
            classGraph.hideOrphanedClasses();

        return classGraph;
    }

    private static URL[] getClasspathUrls(List<File> jarFiles, List<File> classpathDirectories) {
        List<URL> urls = new ArrayList<>();

        classpathDirectories.forEach(classpathDirectory -> {
            try {
                urls.add(classpathDirectory.toURI().toURL());
            } catch (MalformedURLException e) {
                throw new RuntimeException(e);
            }
        });

        jarFiles.forEach((File file) -> {
            try {
                urls.add(file.toURI().toURL());
            } catch (MalformedURLException e) {
                throw new RuntimeException(e);
            }
        });

        return urls.toArray(new URL[urls.size()]);
    }

    private static void addJarToGraph(
            File jarFile, URLClassLoader classLoader, ClassGraph classGraph, CommandlineConfiguration configuration)
            throws IOException, ClassNotFoundException {

        for (String className : getClassNamesFromJar(jarFile))
            attemptClassAdditionByName(classLoader, classGraph, configuration, className);
    }

    private static void attemptClassAdditionByName(URLClassLoader classLoader, ClassGraph classGraph, CommandlineConfiguration configuration, String className) throws ClassNotFoundException {
        if (configuration.isDebug())
            System.out.println("Adding class to graph: " + className);
        try {
            classGraph.add(loadClassByName(classLoader, className));
        } catch (NoClassDefFoundError e) {
            if (configuration.isDebug())
                System.out.println("Class definition was not found.");
            // Sometimes referenced classes are not found in the same Jar.
            // Let's ignore this and proceed with the classes that we have.
        }
    }

    private static Class loadClassByName(URLClassLoader classLoader, String className) throws ClassNotFoundException {
        return Class.forName(className, true, classLoader);
    }

    public static List<String> getClassNamesFromJar(File jarFile) throws IOException {
        List<String> result = new ArrayList<>();
        try (
                JarInputStream jarInputStream = new JarInputStream(new FileInputStream(jarFile))
        ) {
            while (true) {
                JarEntry jarEntry = jarInputStream.getNextJarEntry();
                if (jarEntry == null)
                    break;

                if (isClassFile(jarEntry))
                    result.add(getClassNameFromFileName(jarEntry));
            }

            return result;
        }
    }

    private static boolean isClassFile(JarEntry jarEntry) {
        return jarEntry.getName().endsWith(".class");
    }

    private static String getClassNameFromFileName(JarEntry jarEntry) {
        String result = jarEntry.getName().replaceAll("/", "\\.");
        return result.substring(0, result.lastIndexOf('.'));
    }
}
