/*
 * JavaInspect - Utility to visualize java software
 * Copyright (C) 2013-2020, Svjatoslav Agejenko, svjatoslav@svjatoslav.eu
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 3 of the GNU Lesser General Public License
 * or later as published by the Free Software Foundation.
 */

package eu.svjatoslav.inspector.java.structure;

import eu.svjatoslav.commons.string.GlobMatcher;

import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import static java.io.File.separator;
import static java.lang.System.getProperty;

public class ClassGraph {

    /**
     * Maps class fully qualified names to class descriptors.
     */
    private final Map<String, ClassDescriptor> fullyQualifiedNameToClassMap = new HashMap<>();

    private final List<String> blacklistClassGlobs = new ArrayList<>();

    private final List<String> whitelistClassGlobs = new ArrayList<>();
    TargetImageType targetImageType = TargetImageType.SVG;
    private boolean keepDotFile;

    /**
     * Default to current directory
     */
    private File targetDirectory = new File(getProperty("user.dir") + separator);

    public ClassGraph() {
    }

    /**
     * @param objects objects that shall be added to graph
     * @return this {@link ClassGraph}
     */
    public ClassGraph add(final Object... objects) {

        if (objects != null)
            for (final Object object : objects)
                addObject(object);

        return this;
    }

    private void addObject(final Object object) {
        if (object instanceof Class)
            getOrCreateClassDescriptor((Class) object);
        else
            getOrCreateClassDescriptor(object.getClass());
    }

    public void blacklistClassGlob(final String glob) {
        blacklistClassGlobs.add(glob);
    }

    public void setTargetImageType(TargetImageType targetImageType) {
        this.targetImageType = targetImageType;
    }

    /**
     * @param resultFileName file name for the generated graph. File extension will be
     *                       added automatically. Existing file with the same name will be
     *                       overwritten.
     */

    public void generateGraph(final String resultFileName) {

        final File dotFile = new File(targetDirectory, resultFileName + ".dot");
        final File imageFile = new File(targetDirectory, resultFileName + "." + targetImageType.fileExtension);

        try {
            // write DOT file to disk
            final PrintWriter out = new PrintWriter(dotFile, "UTF-8");
            out.write(getDot());
            out.close();

            // execute GraphViz to visualize graph
            try {
                Runtime.getRuntime()
                        .exec(new String[]{"dot",
                                "-T" + targetImageType.fileExtension,
                                dotFile.getAbsolutePath(),
                                "-o",
                                imageFile.getAbsolutePath()}).waitFor();
            } catch (final InterruptedException ignored) {
            }

            if (!keepDotFile)
                // delete dot file
                if (!dotFile.delete())
                    throw new RuntimeException("Cannot delete file: " + dotFile.getAbsolutePath());

        } catch (final IOException e) {
            throw new RuntimeException("Unable to generate graph: " + e.getMessage(), e);
        }

    }

    private String getDot() {
        final StringBuilder result = new StringBuilder();

        result.append("digraph Java {\n");
        result.append("graph [rankdir=LR, overlap = false, concentrate=true];\n");

        for (final Map.Entry<String, ClassDescriptor> entry : fullyQualifiedNameToClassMap
                .entrySet())
            result.append(entry.getValue().getDot());

        result.append("}\n");

        return result.toString();
    }

    /**
     * @param clazz class that shall be added to graph
     * @return {@link ClassDescriptor} corresponding to given {@link Class}
     */
    protected ClassDescriptor getOrCreateClassDescriptor(final Class clazz) {

        if (clazz == null)
            return null;

        final String classFullyQualifiedName = clazz.getName();

        // reuse existing instance if possible
        if (fullyQualifiedNameToClassMap.containsKey(classFullyQualifiedName))
            return fullyQualifiedNameToClassMap.get(classFullyQualifiedName);

        // create new class descriptor
        final ClassDescriptor newClassDescriptor = new ClassDescriptor(this);
        fullyQualifiedNameToClassMap.put(classFullyQualifiedName,
                newClassDescriptor);

        newClassDescriptor.analyzeClass(clazz);

        return newClassDescriptor;
    }

    /**
     * Hide orphaned class that have no references
     *
     * @return this {@link ClassGraph}
     */
    public ClassGraph hideOrphanedClasses() {

        for (final ClassDescriptor classDescriptor : fullyQualifiedNameToClassMap
                .values())
            classDescriptor.hideClassIfNoReferences();

        return this;
    }

    protected boolean isClassShown(final String className) {
        for (final String pattern : blacklistClassGlobs)
            if (GlobMatcher.match(className, pattern))
                return false;

        if (!whitelistClassGlobs.isEmpty()) {
            for (final String pattern : whitelistClassGlobs)
                if (GlobMatcher.match(className, pattern))
                    return true;
            return false;
        }

        return true;
    }

    public ClassGraph setKeepDotFile(final boolean keepDotFile) {
        this.keepDotFile = keepDotFile;

        return this;
    }

    public ClassGraph setTargetDirectory(File targetDirectory) {
        this.targetDirectory = targetDirectory;
        return this;
    }

    public ClassGraph whitelistClassGlob(final String glob) {
        whitelistClassGlobs.add(glob);
        return this;
    }

}
